/******************************************************************
 * symbolminer.c
 *
 * This file contains routines to find kernel symbols - when they
 * are not exported, of course ...
 *
 * It is possible to find them from the outside (/dev/kmem and friends)
 * but some are only detectable when we are in kernel space. kmalloc
 * is the only symbol we absolutely need to link an LKM into the
 * kernel, and luckily it is relatively easy and reliably to find.
 * All other symbols are found here.
 *
 * Author: greuff <greuff@void.at>
 *
 ********************************************************************/

#define __KERNEL__
#define MODULE
#include "libdis.h"

#define MAX_KERNELSYMS 100

struct kernel_symbol { char *name; unsigned long address; } symtab[MAX_KERNELSYMS];
unsigned int symtab_size=0;

int my_memcmp(const void * cs,const void * ct,char *mask,unsigned int count)
{
   const unsigned char *su1, *su2;
   char *m=mask;
   int res = 0;

   for( su1 = cs, su2 = ct; 0 < count; ++su1, ++su2, count--, ++m)
      if (*m=='1' && (res = *su1 - *su2) != 0)
         break;
   return res;
}

void *my_memmem(void *start, void *cmp, char *cmpmask, unsigned int cmplen, unsigned int count)
{
   unsigned int i;
   void *rval=0;
   for(i=0;i<count && rval==0;i++)
   {
      if(my_memcmp(start,cmp,cmpmask,cmplen)==0)
         rval=start;
      else
         start++;
   }
   return rval;
}

/**
 * strcmp - Compare two strings
 * @cs: One string
 * @ct: Another string
 */
int my_strcmp(const char * cs,const char * ct)
{
   register signed char __res;

   while (1) {
      if ((__res = *cs - *ct++) != 0 || !*cs++)
         break;
   }

   return __res;
}

void print_symbol(int symtabidx)
{
   printk("Found symbol: %s at 0x%08x\n",symtab[symtabidx].name,symtab[symtabidx].address);
}

// attempts to find symbols ...
void gather_symbols(int chatty)
{
   struct instr in;
   unsigned int size;
   disassemble_init(0,INTEL_SYNTAX);

   // startup_32 ... thats easy
   // portability: should work for all kernels
   // improvements: work with ds:0x100000 to determine real
   //               address (but ds is hardcoded so it should work)
   //               see arch/i386/kernel/head.S for details
   symtab[0].name="startup_32";
   symtab[0].address=0xc0100000;
   symtab_size++;
   if(chatty) print_symbol(0);

   // start_kernel
   // the call to start_kernel is in head.S, where startup_32 is too, so
   // do an unsharp pattern matching to find it.
   // portability: should work on all 2.4 kernels, 2.2 and 2.6 have (slightly)
   //              different patterns.
   {
      void *ptr, *ptr_base;
      int found;
      char searchstr[]="\x0f\x01\x15\xFF\xFF\xFF\xFF"
                  "\x0f\x01\x1d\xFF\xFF\xFF\xFF"
                  "\xea\xFF\xFF\xFF\xFF\xFF\xFF"
                  "\xb8\x18\x00\x00\x00\x8e\xd8\x8e\xc0\x8e\xe0\x8e\xe8";

      char searchmask[]="1110000"
                        "1110000"
                        "1000000"
                        "1111111111111";

      ptr=my_memmem((void *)symtab[0].address,searchstr,searchmask,13,0xc0300000-0xc0100000);
      if(!ptr)
      {
         if(chatty) printk("can't find is386 pattern\n");
         return;
      }
      // we know where that pattern is, the next call is a call to start_kernel - use
      // the disassembler, Luke
      // TODO: when CONFIG_SMP is defined, it's the second call.
      ptr_base=ptr;
      found=0;
      while(found==0 && ptr<ptr_base+500)   // should well be in the next 500 bytes
      {
         size=disassemble_address(ptr,&in);
         if(size)
         {
            if(!my_strcmp(in.mnemonic,"call"))
            {
               found=1;
            }
            else
               ptr+=size;
         }
         else
         {
            if(chatty) printk("Choking at 0x%08x!\n",ptr);
            ptr++;
         }
      }
      if(found==0)
      {
         if(chatty) printk("can't find call to start_kernel\n");
         return;
      }
      symtab[1].name="start_kernel";
      symtab[1].address=(unsigned long)ptr+(*(unsigned long *)(ptr+1))+size;
      if(chatty) print_symbol(1);
   }
   disassemble_cleanup();
}

int init_module(void)
{
   gather_symbols(1);
}

int cleanup_module(void)
{
}
